using System.Collections;
using System.Collections.Generic;
using UnityEngine;
using System.Linq;
using System.IO;

public class TSPsolver : MonoBehaviour
{
    public List<GameObject> CitiesList;

    public List<List<float>> DistanceList;
    List<List<float>> PheromoneList;
    public List<List<float>> ProbList;
    List<List<GameObject>> PathList;
    List<int> ShortestPath;

    [Header("Prefabs"), SerializeField]
    public GameObject PathPrefab;
    public GameObject CityPrefab;


    private float CurrentTime { get; set; }
    [Header("Settings"), SerializeField] private float updateSpan = 0.1f;
    public float alpha;
    public float beta;
    public float evaporate = 0.02f;
    public float Q = 1;
    public int InitialCityNum;
    private float UpdateSpan => updateSpan;
    public float ColorIntensity = 1.0f;
    float Width;

    [Header("Data"), SerializeField]
    public int StepNum;
    public float Shortest;

    void Start()
    {
        CurrentTime = 0.0f;
        StepNum = 0;
        // 都市をInitialCityNumだけランダムな位置に生成する
        for(int i = 0; i < InitialCityNum; i++)
        {
            GameObject city = Instantiate(CityPrefab) as GameObject;
            Vector3 pos = new Vector3(Random.Range(-7.0f, 7.0f), Random.Range(-4.0f, 4.0f), 0);
            city.transform.position = pos;
            CityController cityController = city.GetComponent<CityController>();
            cityController.SetCityId(i);
        }
        this.CitiesList = GameObject.FindGameObjectsWithTag("City").ToList();
        
        this.DistanceList = new List<List<float>>();
        this.PheromoneList = new List<List<float>>();
        this.ProbList = new List<List<float>>();
        AgentGenerator agentgenerator = GameObject.Find("AgentGenerator").GetComponent<AgentGenerator>();
        this.Width = 0.05f;
        this.Shortest = 10000;
        this.PathList = new List<List<GameObject>>();
        
        // 各都市間の処理
        for(int i = 0; i < CitiesList.Count;i++)
        {
            this.DistanceList.Add(new List<float>());
            this.PheromoneList.Add(new List<float>());
            this.ProbList.Add(new List<float>());
            this.PathList.Add(new List<GameObject>());
            for(int j = 0;j < CitiesList.Count;j++)
            {
                this.DistanceList[i].Add(Vector3.Distance(CitiesList[i].transform.position, CitiesList[j].transform.position));
                this.PheromoneList[i].Add(1);
                this.ProbList[i].Add(1);
                this.PathList[i].Add(null);
            }
        }
        for(int i = 0; i < CitiesList.Count; i++)
        {
            for(int j = 0; j < i; j++)
            {
                GameObject path = Instantiate(PathPrefab) as GameObject;
                var linerenderer = path.GetComponent<LineRenderer>();
                var pathpos = new Vector3[]{CitiesList[i].transform.position, CitiesList[j].transform.position};

                linerenderer.material = new Material(Shader.Find("Sprites/Default"));
                linerenderer.SetPositions(pathpos);
                linerenderer.startWidth = Width;
                linerenderer.endWidth = Width;
                linerenderer.startColor = new Color(0f,1f,0f,1f);
                linerenderer.endColor = new Color(0f,1f,0f,1f);
                this.PathList[i][j] = this.PathList[j][i] = path;
            }
        }
    }

    void FixedUpdate()
    {
        // 各都市間についてフェロモンの蒸発、遷移確率の更新、pathの更新
        // 毎フレーム更新すると重いのでUpdateSpanごとに
        CurrentTime += Time.deltaTime;
        if(CurrentTime > UpdateSpan) {
            StepNum++;
            for(int i = 0; i < CitiesList.Count; i++)
            {
                for(int j = 0; j < CitiesList.Count; j++)
                {
                    if(i == j)continue;
                    this.PheromoneList[i][j] *= (1 - evaporate);
                    this.ProbList[i][j] =  Mathf.Pow(PheromoneList[i][j], alpha) * Mathf.Pow(DistanceList[i][j], -beta);

                    var linerenderer = this.PathList[i][j].GetComponent<LineRenderer>();
                    double coloralpha = System.Math.Tanh(this.PheromoneList[i][j]/100*ColorIntensity);
                    Color newcolor = linerenderer.startColor;
                    newcolor.a = (float)coloralpha;
                    linerenderer.startColor = newcolor;
                    linerenderer.endColor = newcolor;

                }
            }
            CurrentTime = 0.0f;
        }
    }

    public void RenewDistanceList(GameObject city)
    {
        // 対象の都市について距離の更新
        for(int i = 0; i < CitiesList.Count; i++)
        {
            if(CitiesList[i].GetInstanceID() == city.GetInstanceID())
            {
                CitiesList[i] = city;
                for(int j = 0; j < CitiesList.Count; j++)
                {
                    if(i == j)continue;
                    this.DistanceList[i][j] = Vector3.Distance(CitiesList[i].transform.position, CitiesList[j].transform.position);
                    this.DistanceList[j][i] = this.DistanceList[i][j];
                    var linerenderer = this.PathList[i][j].GetComponent<LineRenderer>();
                    var pathpos = new Vector3[]{CitiesList[i].transform.position, CitiesList[j].transform.position};
                    linerenderer.SetPositions(pathpos);
                }
            }
        }

        // 最短距離の更新
        if(ShortestPath != null){
            float NewShortest = 0;
            for(int i = 0; i < this.ShortestPath.Count; i++)
            {
                int id1 = ShortestPath[i];
                int id2 = (i == this.ShortestPath.Count-1) ? ShortestPath[0] : ShortestPath[i+1];
                NewShortest += this.DistanceList[id1][id2];
            }
            this.Shortest = NewShortest;
        }
    }
    public void RenewPheromoneList(int i, int j)
    {
        PheromoneList[i][j] += Q;
        PheromoneList[j][i] += Q;
    }

    public void AddCity(GameObject city)
    {
        city.GetComponent<CityController>().SetCityId(CitiesList.Count);
        this.CitiesList.Add(city);
        this.DistanceList.Add(new List<float>());
        this.PheromoneList.Add(new List<float>());
        this.ProbList.Add(new List<float>());
        this.PathList.Add(new List<GameObject>());
        
        // cityの情報を格納するための要素を追加
        for(int i = 0;i < CitiesList.Count-1;i++)
        {
            this.DistanceList[i].Add(Vector3.Distance(CitiesList[i].transform.position, city.transform.position));
            this.PheromoneList[i].Add(1);
            this.ProbList[i].Add(1);
            this.PathList[i].Add(null);
        }

        for(int j = 0;j < CitiesList.Count;j++)
        {
            this.DistanceList[CitiesList.Count-1].Add(Vector3.Distance(CitiesList[j].transform.position, city.transform.position));
            this.PheromoneList[CitiesList.Count-1].Add(1);
            this.ProbList[CitiesList.Count-1].Add(1);
            this.PathList[CitiesList.Count-1].Add(null);
        }

        // 必要な各要素を追加
        for(int i = 0; i < CitiesList.Count-1; i++)
        {
            GameObject path = Instantiate(PathPrefab) as GameObject;
            var linerenderer = path.GetComponent<LineRenderer>();
            var pathpos = new Vector3[]{CitiesList[i].transform.position, city.transform.position};
            linerenderer.material = new Material(Shader.Find("Sprites/Default"));
            linerenderer.SetPositions(pathpos);
            linerenderer.startWidth = Width;
            linerenderer.endWidth = Width;
            linerenderer.startColor = new Color(0f,1f,0f,1f);
            linerenderer.endColor = new Color(0f,1f,0f,1f);
            this.PathList[CitiesList.Count-1][i] = this.PathList[i][CitiesList.Count-1] = path;
        }
        this.DestroyAgent();
        this.Shortest = 10000;
    }
    public void RenewShortestPath(List<int> VisitedList)
    {
        // 最短距離の更新時にpathの色を変更する
        
        // 古い最短距離のpathの色をデフォルトにする
        if(this.ShortestPath != null)
        {
            for(int i = 0; i < this.ShortestPath.Count; i++)
            {
                GameObject path;
                int id1 = ShortestPath[i];
                int id2 = (i == this.ShortestPath.Count-1) ? ShortestPath[0] : ShortestPath[i+1];
                path = this.PathList[id1][id2];
                double coloralpha = System.Math.Tanh(this.PheromoneList[id1][id2]/100*ColorIntensity);
                
                var linerenderer = path.GetComponent<LineRenderer>();
                linerenderer.material = new Material(Shader.Find("Sprites/Default"));
                linerenderer.startColor = new Color(0f,1f,0f,(float)coloralpha);
                linerenderer.endColor = new Color(0f,1f,0f,(float)coloralpha);
                linerenderer.sortingOrder = 0;
            }
        }
        // 新しい最短距離のpathの色を変える
        for(int i = 0; i < VisitedList.Count; i++)
        {
            GameObject newpath;
            int id1 = VisitedList[i];
            int id2 = (i == VisitedList.Count-1) ? VisitedList[0] : VisitedList[i+1];
            // Debug.Log(id1.ToString() + " " + id2.ToString());
            newpath = this.PathList[id1][id2];
            double coloralpha = System.Math.Tanh(this.PheromoneList[id1][id2]/100*ColorIntensity);
            
            var linerenderer = newpath.GetComponent<LineRenderer>();
            linerenderer.material = new Material(Shader.Find("Sprites/Default"));
            linerenderer.startColor = new Color(1f,0f,1f,(float)coloralpha);
            linerenderer.endColor = new Color(1f,0f,1f,(float)coloralpha);
            linerenderer.sortingOrder = 1;
        }
        
        this.ShortestPath = VisitedList;
        // Debug.Log(string.Join(",",ShortestPath));
    }
    void Reset()
    {
        this.DestroyCity();
        this.DestroyAgent();
        this.DestroyPath();
        this.Start();
    }
    void DestroyCity()
    {
        for(int i = 0; i < this.CitiesList.Count; i++)
        {
            DestroyImmediate(this.CitiesList[i]);
        }
    }
    void DestroyAgent()
    {
        List<GameObject> agentlist = GameObject.FindGameObjectsWithTag("Agent").ToList();
        for(int i = 0; i < agentlist.Count; i++)
        {
            DestroyImmediate(agentlist[i]);
        }
    }
    void DestroyPath()
    {
        for(int i = 0; i < this.PathList.Count; i++)
        {
            for(int j = 0; j < i; j++)
            {
                DestroyImmediate(this.PathList[i][j]);
            }
        }
    }
}
